import random
import string


class DataNotFound(IndexError):
    pass


def convert_type(x):
    if str(x).endswith('%'):
        return float(x[:-1].replace('.', '').replace(',', '.')) / 100
    try:
        if float(x).is_integer():
            return int(float(x))
        return float(x)
    except ValueError:
        return None


def get_headers():
    return {
        'User-Agent': ''.join(
            random.choices(string.ascii_letters, k=10)
        )
    }



import requests
import xlrd
import pandas as pd
import io
import os
from zipfile import ZipFile
from .utils import convert_type, DataNotFound, get_headers


def get_tickers():
    '''Downloads tickers' codes from fundamentus website

    :Returns:

    A pandas DataFrame with three columns:

    Papel (Ticker),
    Nome Comercial (Trade Name),
    Razão Social (Corporate Name)
    '''
    html_src = requests.get(
        'http://fundamentus.com.br/detalhes.php', headers=get_headers()).text
    return pd.read_html(html_src)[0]


def _get_sheets(ticker, quarterly, ascending):
    '''Downloads sheets from fundamentus website.

    available sheets:
        \'Bal. Patrim.\' for balance sheet
        \'Dem. Result.\' for income statement
    '''

    ticker = ticker.upper()

    # Apparently fundamentus is blocking requests library's standard user-agent
    # To solve this problem, I'm generating random user-agents

    r = requests.get('https://www.fundamentus.com.br/balancos.php',
                     params={'papel': ticker},
                     headers=get_headers())

    SID = r.cookies.values()[0]

    response_sheet = requests.get(
        'https://www.fundamentus.com.br/planilhas.php',
        params={'SID': SID}, headers=get_headers())

    if response_sheet.text.startswith('Ativo nao encontrado'):
        raise DataNotFound(
            f'Couldn\'t find any data for {ticker}')

    with io.BytesIO(response_sheet.content) as zip_bytes:
        with ZipFile(zip_bytes) as z:
            xls_bytes = z.read('balanco.xls')

    # Supress warnings
    wb = xlrd.open_workbook(file_contents=xls_bytes,
                            logfile=open(os.devnull, 'w'))

    dfs = {
        'Bal. Patrim.': pd.read_excel(wb, engine='xlrd',
                                      index_col=0,
                                      sheet_name='Bal. Patrim.'),
        'Dem. Result.': pd.read_excel(wb, engine='xlrd',
                                      index_col=0,
                                      sheet_name='Dem. Result.')
    }

    for sheet, df in dfs.items():

        # Cleaning the DataFrame
        df.columns = df.iloc[0, :]
        df = df.iloc[1:].T.applymap(convert_type)
        df.index.name = 'Data'
        df.columns.name = ticker
        df = df.loc[df.index.notnull()]
        df.index = pd.to_datetime(
            df.index, format='%d/%m/%Y')

        # Excluding empty columns (usually subtitles)
        df.dropna(axis=1, how='all', inplace=True)
        # Filling empty cells with 0
        df.fillna(0, inplace=True)

        if not quarterly:
            rows_to_drop = [x for x in df.index.year
                            if list(df.index.year).count(x) != 4]
            df = df.groupby(df.index.year).sum()
            df.drop(rows_to_drop, inplace=True)
            df.index = [str(x) for x in df.index]

        dfs[sheet] = df.sort_index(ascending=ascending).astype(int)

    return dfs


def get_balanco(ticker, quarterly=False, ascending=True, separated=True):
    '''Get the balance sheet for one brazilian stock listed on fundamentus website.

    NOTE: Values are in thousands!

    :Arguments:

    ticker[str]:
        Ticker to download data from
    quarterly[bool]:
        Whether to download quarterly or annualy data.
        Default is False.
    ascending[bool]:
        Whether the date index should be sorted ascendingly on the DataFrame
        Default is True
    separated[bool]:
        If True, the DataFrame will be hierarchically divided by super columns:
            \'Ativo Total\' (Total Assets),
            \'Ativo Circulante\' (Current Assets),
            \'Ativo Não Circulante\' (Non-current Assets),
            \'Passivo Total\' (Total Liabilities),
            \'Passivo Circulante\' (Current Liabilities),
            \'Passivo Não Circulante\' (Non-current Liabilities),
            \'Patrimônio Líquido\' (Net Worth)
        Default is True (highly recommended, as some infra columns are
        duplicated which could lead to confusion).

    :Raises:

    DataNotFound(IndexError) if there's no data available for that ticker

    :Returns:

    A pandas DataFrame
    '''

    df = _get_sheets(ticker, quarterly=quarterly,
                     ascending=ascending)['Bal. Patrim.']

    if separated:
        super_cols = [
            'Ativo Total',
            'Ativo Circulante',
            'Ativo Não Circulante',
            'Ativo Realizável a Longo Prazo',
            'Passivo Total',
            'Passivo Circulante',
            'Passivo Não Circulante',
            'Passivo Exigível a Longo Prazo',
            'Patrimônio Líquido'
        ]

        cols = list(df.columns)

        # Handling different balance sheets for banks and other companies
        if 'Ativo Não Circulante' in cols:
            super_cols.remove('Ativo Realizável a Longo Prazo')
        else:
            super_cols.remove('Ativo Não Circulante')

        if 'Passivo Não Circulante' in cols:
            super_cols.remove('Passivo Exigível a Longo Prazo')
        else:
            super_cols.remove('Passivo Não Circulante')

        idxs = [cols.index(x) for x in cols if x in super_cols]

        slices = [slice(idxs[i], idxs[i + 1])
                  if i < (len(idxs) - 1)
                  else slice(idxs[i], None)
                  for i, _ in enumerate(idxs)]

        tuples = []

        for s in slices:
            sup = super_cols.pop(0)

            # Renaming columns to standardize different companies' DataFrames
            if sup == 'Ativo Realizável a Longo Prazo':
                sup = 'Ativo Não Circulante'
            if sup == 'Passivo Exigível a Longo Prazo':
                sup = 'Passivo Não Circulante'

            for col in cols[s]:
                tuples.append((sup, col))

        df.columns = pd.MultiIndex.from_tuples(tuples)

    return df


def get_dre(ticker, quarterly=False, ascending=True):
    '''Get the income statement for one brazilian stock listed on fundamentus website.

    NOTE: Values are in thousands!

    :Arguments:

    ticker[str]:
        Ticker to download data from
    quarterly[bool]:
        Whether to download quarterly or annualy data.
        Default is False.
    ascending[bool]:
        Whether the date index should be sorted ascendingly on the DataFrame
        Default is True

    :Raises:

    DataNotFound(IndexError) if there's no data available for that ticker

    :Returns:

    A pandas DataFrame
    '''

    df = _get_sheets(ticker, quarterly=quarterly,
                     ascending=ascending)['Dem. Result.']

    return df



from .advfn import get_fundamentos
from .advfn import get_schema
from .fundamentus import get_balanco
from .fundamentus import get_dre
from .fundamentus import get_tickers
from .utils import DataNotFound



import requests
import pandas as pd
from .utils import convert_type, DataNotFound, get_headers
from concurrent import futures
from datetime import date
from calendar import monthrange


schema = {
    'Quantidade de Ações ON': 'Quantidade de Ações ON',
    'Quantidade de Ações PN': 'Quantidade de Ações PN',
    'Última Cotação ON': 'Última Cotação ON',
    'Última Cotação PN': 'Última Cotação PN',
    'Valor de Mercado': 'Valor de Mercado',
    'Receita Líquida': 'Receita Líquida',
    'Price Sales Ratio (PSR)': 'PSR',
    'Resultado Bruto': 'Resultado Bruto',
    'Margem Bruta': 'Margem Bruta',
    'EBITDA': 'EBITDA',
    'Margem EBITDA': 'Margem EBITDA',
    'Preço / EBITDA': 'P/EBITDA',
    'EBIT': 'EBIT',
    'Margem EBIT': 'Margem EBIT',
    'Preço / EBIT': 'P/EBIT',
    'Lucro/Prejuízo Líquido': 'Lucro Líquido',
    'Preço / Lucro (P/L)': 'P/L',
    'Lucro por Ação (LPA)': 'LPA',
    'Ativo Total': 'Ativo Total',
    'Preço / Ativo (P/A)': 'P/Ativo',
    'Giro Ativos': 'Giro',
    'EBIT / Ativo': 'EBIT/Ativo',
    'Patrimônio Líquido': 'PL',
    'Retorno sobre PL (ROE)': 'ROE',
    'Valor Patrimonial por Ação (VPA)': 'VPA',
    'Preço / Valor Patrimonial por Ação (P/VPA)': 'P/VPA',
    '(Ativo - Patrimônio Líquido) / Patrimônio Líquido': '(A - PL)/PL',
    'Equity Multiplier (EM)': 'EM',
    'Disponibilidades': 'Disponibilidades',
    'Dinheiro em Caixa': 'Dinheiro em Caixa',
    'Aplicações Financeiras': 'Aplicações Financeiras',
    'Dívida Bruta': 'Dívida Bruta',
    'Dívida Bruta / Patrimônio Líquido': 'DB/PL',
    'Endividamento Financeiro': 'Endividamento Financeiro',
    'Dívida Líquida': 'Dívida Líquida',
    'Dívida Líquida / EBITDA': 'DL/EBITDA',
    'Enterprise Value (EV)': 'EV',
    'Enterprise Value / EBIT (EV/EBIT)': 'EV/EBIT',
    'Ativo Circulante': 'Ativo Circulante',
    'Ativo Não Circulante': 'Ativo Não Circulante',
    'Ativo Circulante Líquido': 'Ativo Circulante Líquido',
    'Preço / Ativo Circulante Líquido': 'P/ACL',
    'Passivo Circulante': 'Passivo Circulante',
    'Passivo Não Circulante': 'Passivo Não Circulante',
    'Liquidez Corrente': 'LC',
    'Liquidez Imediata': 'LI',
    'Capital de Giro': 'Capital de Giro',
    'Preço / Capital de Giro': 'P/Capital de Giro',
    'Dívida em Moeda Estrangeira': 'Dívida em Moeda Estrangeira',
    'Fluxo de Caixa Operacional (FCO)': 'FCO',
    'Fluxo de Caixa de Investimentos (FCI)': 'FCI',
    'Fluxo de Caixa de Financiamentos (FCF)': 'FCF',
    'Fluxo de Caixa Total (FCT)': 'FCT',
    'Fluxo de Caixa Livre (FCL)': 'FCL',
    'CAPEX': 'CAPEX',
    'Fluxo de Caixa Livre CAPEX': 'FCL CAPEX',
    'CAPEX / Fluxo de Caixa Operacional': 'CAPEX/FCO',
    'Fluxo de Caixa de Investimentos / Lucro Líquido': 'FCI/LL',
    'CAPEX / Lucro Líquido': 'CAPEX/LL',
    'Dividendos e Juros Sobre Capital Próprio Pagos': 'Dividendos e JCP',
    'Dividend Yield': 'DY',
    'Dividend Payout': 'Payout'
}


def get_schema():
    '''Get the schema used to abbreviate columns names on get_fundamentos DataFrame
    '''
    return (pd.DataFrame(schema.items(), columns=['Significado', 'Abreviação'])
            .set_index('Abreviação'))


def get_fundamentos(ticker, year=None, quarter=None,
                    separated=True, ascending=True, threads=True):
    '''Get fundamental data for one brazilian stock listed on ADVFN website.

    NOTE: Values are in thousands!

    Also, none of the data is in percentage scale.

    :Arguments:

    ticker[str]:
        Ticker to download data from
    year[int] (2007 - today):
        Year to download data from. If None, downloads maximum range.
        Default is None
    quarter[int] (1 - 4):
        Quarter to download data from. If None, downloads annual data.
        Default is None.
    separated[bool]:
        If True, the DataFrame will be hierarchically divided by super columns.
        Each super column is a different topic of economic indicators
            \'Mercado\' (Market),
            \'Resultados\' (Income),
            \'Patrimônio\'(Net Worth),
            \'Caixa\' (Cash),
            \'Dívida\' (Debt),
            \'Liquidez e Solvência\' (Solvency and Liquidity),
            \'Fluxo de Caixa\' (Cash Flow),
            \'Investimentos\' (Investments),
            \'Dividendos\' (Dividends)
        Default is True
    ascending[bool]:
        If downloading multiple years, whether they should be sorted ascendingly
         on the DataFrame
        Default is True
    threads[bool]:
        If downloading multiple years, whether to download data using multiple
         threads.
        Default is True (highly recommended)

    :Raises:

    ValueError if the argument year is invalid
    DataNotFound(IndexError) if there's no data available for that specific
     ticker or year

    :Returns:

    A pandas DataFrame, if dfs = False
    '''
    ticker = ticker.upper()

    if year is None:
        years = [x for x in range(2007, date.today().year + 1)]
        annual_dfs = []

        # Recursive search
        if threads:

            futures_list = []
            with futures.ThreadPoolExecutor(max_workers=15) as executor:
                for year in years:
                    future = executor.submit(
                        get_fundamentos, ticker, year, quarter, separated)
                    futures_list.append(future)
            for future in futures_list:
                try:
                    annual_dfs.append(future.result())
                except (DataNotFound, ValueError):
                    continue
        else:
            for year in years:
                try:
                    annual_dfs.append(get_fundamentos(
                        ticker, year, quarter, separated))
                except (DataNotFound, ValueError):
                    continue

        if not annual_dfs:
            raise DataNotFound(f'Couldn\'t find any data for \'{ticker}\'')

        df = pd.concat(annual_dfs)

        return df.sort_index(ascending=ascending)

    if year > int(date.today().year):
        raise ValueError(
            f'Year variable ca\'nt be higher than current year: {year}')

    str_tri = {
        1: 'primeiro-trimestre', 2: 'segundo-trimestre',
        3: 'terceiro-trimestre', 4: 'quarto-trimestre',
        None: None
    }

    baseurl = 'https://br.advfn.com/bolsa-de-valores/bovespa/{0}/fundamentos/individualizado/{1}/{2}'
    html_src = requests.get(baseurl.format(
        ticker, year, str_tri[quarter]), headers=get_headers()).text

    _dfs = pd.read_html(html_src, index_col=0, decimal=',', thousands='.')

    if len(_dfs) < 9:
        raise DataNotFound(
            f'Couldn\'t find any data for \'{ticker}\' on {year}')

    _dfs = _dfs[:9]

    for i, df in enumerate(_dfs):
        # Formating types and dropping entire NaN columns
        _dfs[i] = df.iloc[:, [0]].T.applymap(convert_type)
        _dfs[i].dropna(how='all', axis=1, inplace=True)

        if quarter:
            # Formating index column to be datetime
            _dfs[i].index = [f'{monthrange(year, quarter*3)[1]}/{quarter*3}/{x[3:]}'
                             for x in _dfs[i].index]
            _dfs[i].index = pd.to_datetime(_dfs[i].index,
                                           format='%d/%m/%Y')
        else:
            # Removing '*' from some years
            _dfs[i].index = [year if '*' not in year else year[:-2]
                             for year in _dfs[i].index]

        _dfs[i].columns = [schema[x] for x in _dfs[i].columns]
        _dfs[i].columns.name = ticker
        _dfs[i].index.name = 'Data'

    if separated:
        super_cols = [
            'Mercado', 'Resultados', 'Patrimônio', 'Caixa', 'Dívida',
            'Liquidez e Solvência', 'Fluxo de Caixa', 'Investimentos',
            'Dividendos',
        ]
        return pd.concat(_dfs, axis=1, keys=super_cols)

    return pd.concat(_dfs, axis=1)



from setuptools import setup

# twine check dist/* pra ver se n tem nada de errado
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='fundamentos',
    packages=['fundamentos'],
    version='1.5',  # Atualizar
    license='MIT',
    description='Download Bovespa Stock Market fundamentals with Python.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Nathan de Moura Vieira',
    author_email='nathanmoura194@gmail.com',
    url='https://github.com/NathanMoura/fundamentos',
    download_url='https://github.com/NathanMoura/fundamentos/archive/v1.5.tar.gz',  # Atualizar
    keywords=['pandas', 'finance', 'fundamentals', 'bovespa'],
    install_requires=[
        'pandas',
        'requests',
        'lxml',
        'xlrd'
    ],
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 5 - Production/Stable',

        'Operating System :: OS Independent',

        'Intended Audience :: Developers',

        'Topic :: Office/Business :: Financial',
        'Topic :: Office/Business :: Financial :: Investment',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',

    ],
)
